#!/bin/sh
#
# A tool to simplify Makefiles that need to put something
# into the ROMFS
#
#############################################################################

# Provide a default PATH setting to avoid potential problems...
PATH="/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:$PATH"

usage()
{
cat << !EOF >&2
$0: [options] [src] dst
    -v          : output actions performed.
    -e env-var  : only take action if env-var is set to "y".
    -o option   : only take action if option is set to "y".
    -p perms    : chmod style permissions for dst.
    -a text     : append text to dst.
    -l link     : dst is a link to 'link'.
    -s sym-link : dst is a sym-link to 'sym-link'.

    if "src" is not provided,  basename is run on dst to determine the
    source in the current directory.

	multiple -e and -o options are ANDed together.  To achieve an OR affect
	use a single -e/-o with 1 or more y/n/"" chars in the condition.

	if src is a directory,  everything in it is copied recursively to dst
	with special files removed (currently CVS dirs).
!EOF
	exit 1
}

#############################################################################

setperm()
{
	if [ "$perm" ]
	then
		[ "$v" ] && echo "chmod ${perm} ${TARGET}${dst}"
		chmod ${perm} ${TARGET}${dst}
	fi
}

#############################################################################

file_copy()
{
	if [ -d "${src}" ]
	then
		[ "$v" ] && echo "CopyDir ${src} ${TARGET}${dst}"
		(
			cd ${src}
			V=
			[ "$v" ] && V=v
			find . -print | grep -E -v '/CVS' | cpio -p${V}dumL ${TARGET}${dst}
		)
	else
		rm -f ${TARGET}${dst}
		[ "$v" ] && echo "cp ${src} ${TARGET}${dst}"
		cp ${src} ${TARGET}${dst}
		setperm ${TARGET}${dst}
	fi
}

#############################################################################

file_append()
{
	touch ${TARGET}${dst}
	if grep -F "${src}" ${TARGET}${dst}
	then
		[ "$v" ] && echo "File entry already installed."
	else
		[ "$v" ] && echo "Installing entry into ${TARGET}${dst}."
		echo "${src}" >> ${TARGET}${dst}
	fi
	setperm ${TARGET}${dst}
}

#############################################################################

hard_link()
{
	rm -f ${TARGET}${dst}
	[ "$v" ] && echo "ln ${src} ${TARGET}${dst}"
	ln ${src} ${TARGET}${dst}
}

#############################################################################

sym_link()
{
	rm -f ${TARGET}${dst}
	[ "$v" ] && echo "ln -s ${src} ${TARGET}${dst}"
	ln -s ${src} ${TARGET}${dst}
}

#############################################################################
#
# main program entry point
#

if [ -z "$TARGET" ]
then
	echo "TARGET is not set" >&2
	exit 1
fi

v=
option=y
perm=
func=file_copy
src=
dst=

while getopts 've:o:p:a:l:s:' opt "$@"
do
	case "$opt" in
	v) v="1";                           ;;
	o) option="$OPTARG";                ;;
	e) eval option=\"\$$OPTARG\";       ;;
	p) perm="$OPTARG";                  ;;
	a) src="$OPTARG"; func=file_append; ;;
	l) src="$OPTARG"; func=hard_link;   ;;
	s) src="$OPTARG"; func=sym_link;    ;;

	*)  break ;;
	esac
#
#	process option here to get an ANDing effect
#
	case "$option" in
	*[yY]*) # this gives OR effect, ie., nYn
		;;
	*)
		[ "$v" ] && echo "Condition not satisfied."
		exit 0
		;;
	esac
done

shift `expr $OPTIND - 1`

case $# in
1)
	dst="$1"
	if [ -z "$src" ]
	then
		src="`basename $dst`"
	fi
	;;
2)
	if [ ! -z "$src" ]
	then
		echo "Source file already provided" >&2
		exit 1
	fi
	src="$1"
	dst="$2"
	;;
*)
	usage
	;;
esac

$func

exit 0

#############################################################################
